__version__ 	= "1.0"
__author__ 		= ['Ricardo Ribeiro', 'Sergio Copeto']
__credits__ 	= ['Ricardo Ribeiro', "Sergio Copeto"]
__license__ 	= "Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>"
__maintainer__ 	= ['Ricardo Ribeiro', 'Sergio Copeto']
__email__ 		= ['ricardo.ribeiro@research.fchampalimaud.org', 'sergio.copeto']
__status__ 		= "Development"

from confapp import conf

conf += 'pybpodgui_plugin_waveplayer.settings'

from pybpodgui_plugin_waveplayer.module import WavePlayer as BpodModule


# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging

from confapp import conf


from AnyQt.QtGui import QIcon
from AnyQt import QtCore

from pybpodgui_plugin.models.session.session_window import SessionWindow

logger = logging.getLogger(__name__)


class SessionTreeNode(SessionWindow):
    def __init__(self, setup):
        SessionWindow.__init__(self, setup)

        self.__running_icon = QIcon(conf.PLAY_SMALL_ICON)
        # this helps on cascade elimination of treenodes
        self.subjects_nodes = {}
        self.running = False
        self.create_treenode(self.tree)

    def create_treenode(self, tree):
        self.node = tree.create_child(self.name, self.setup.node)
        self.node.key_pressed_event = self.node_key_pressed_event
        self.node.double_clicked_event = self.node_double_clicked_event
        self.node.window = self
        self.node.setExpanded(True)

        tree.add_popup_menu_option('Remove', self.remove, item=self.node, icon=QIcon(conf.REMOVE_SMALL_ICON))

        return self.node

    def node_key_pressed_event(self, event):
        if event.key() == QtCore.Qt.Key_Delete:
            self.remove()

    def node_double_clicked_event(self):
        self.load_contents()

    @property
    def name(self):
        if hasattr(self, 'node'):
            return str(self.node.text(0))
        else:
            return SessionWindow.name.fget(self)

    @name.setter
    def name(self, value):
        SessionWindow.name.fset(self, value)
        if hasattr(self, 'node'):
            self.node.setText(0, value)

    @property
    def tree(self):
        return self.setup.tree

    ##########################################################################
    ####### FUNCTIONS ########################################################
    ##########################################################################

    def remove(self):
        if self.setup.MARKED_FOR_REMOVAL:
            reply = 'yes'
        else:
            reply = self.question('Delete session: ' + self.name + '?', 'Delete')

        if reply == 'yes':
            super(SessionTreeNode, self).remove()
            self.setup.node.removeChild(self.node)



import datetime, json, uuid
from sca import format_flags as flags
from sca.utils import hash_utils
from dateutil import parser


class scadict(dict):

    def __init__(self, data=None, uuid4_id=None, def_url=None, def_text=None, software=None):
        self.uuid4      = str(uuid.uuid4() if uuid4_id is None else uuid4_id) 
        self.def_url    = def_url 
        self.def_text   = def_text 
        self.software   = software

        if data: self.update(data)
            
        self.updated = False

        
    def __setitem__(self, key, value):
        if self.get(key, None)!=value: self.updated = True
        super(scadict, self).__setitem__(key, value)

    @property
    def created_on(self): 
        return self.get(flags.CREATED_ON_FLAG, None)
    @created_on.setter
    def created_on(self, value):
        if self.get(flags.CREATED_ON_FLAG, None)!=value: self.updated = True
        
        self[flags.CREATED_ON_FLAG] = value.isoformat(' ') if value else value

    @property
    def updated_on(self): 
        return self.get(flags.UPDATED_ON_FLAG, None)
    @updated_on.setter
    def updated_on(self, value):
        if self.get(flags.UPDATED_ON_FLAG, None)!=value: self.updated = True
        
        self[flags.UPDATED_ON_FLAG] = value.isoformat(' ') if value else value

    @property
    def uuid4(self): 
        return self.get(flags.UUID4_FLAG, None)
    @uuid4.setter
    def uuid4(self, value):
        if self.get(flags.UUID4_FLAG, None)!=str(value): self.updated = True
        
        self[flags.UUID4_FLAG] = str(value)

    @property
    def def_url(self): 
        return self.get(flags.DEFINITION_URL_FLAG, None)
    @def_url.setter
    def def_url(self, value):
        if self.get(flags.DEFINITION_URL_FLAG, None)!=value: self.updated = True
        
        self[flags.DEFINITION_URL_FLAG] = value

    @property
    def def_text(self): 
        return self.get(flags.DEFINITION_TEXT_FLAG, None)
    @def_text.setter
    def def_text(self, value):
        if self.get(flags.DEFINITION_TEXT_FLAG, None)!=value:
            self.updated = True
        self[flags.DEFINITION_TEXT_FLAG] = value

    @property
    def software(self): 
        return self.get(flags.SOFTWARE_REF_FLAG, None)
    @software.setter
    def software(self, value):
        if self.get(flags.SOFTWARE_REF_FLAG, None)!=value:
            self.updated = True
        self[flags.SOFTWARE_REF_FLAG] = value


    def add_parent_ref(self, ref_uuid4): 
        if flags.PARENT_REF_FLAG not in self:
            self[flags.PARENT_REF_FLAG] = []
            self.updated = True
        if str(ref_uuid4) not in self[flags.PARENT_REF_FLAG]:
            self[flags.PARENT_REF_FLAG].append(str(ref_uuid4))
            self.updated = True
 
    def add_external_ref(self, ref_uuid4):
        if flags.EXTERNAL_REF_FLAG not in self:
            self[flags.EXTERNAL_REF_FLAG] = []
            self.updated = True

        if str(ref_uuid4) not in self[flags.EXTERNAL_REF_FLAG]:
            self[flags.EXTERNAL_REF_FLAG].append(str(ref_uuid4)) 
            self.updated = True
     
    def add_external_url(self, url): 
        if flags.EXTERNAL_URL_FLAG not in self:
            self[flags.EXTERNAL_URL_FLAG] = [] 
            self.updated = True

        if url not in self[flags.EXTERNAL_URL_FLAG]:
            self[flags.EXTERNAL_URL_FLAG].append(url) 
            self.updated = True
 
    def add_external_file(self, filepath): 
        if flags.EXTERNAL_FILE_FLAG not in self:
            self[flags.EXTERNAL_FILE_FLAG] = [] 
            self.updated = True

        filename, hash_algorithm, hash_value = hash_utils.calculate_hash(filepath)

        found = False
        for filedata in self.get(flags.EXTERNAL_FILE_FLAG, []):
            f  = filedata.get('filename', None)
            h  = filedata.get('hash-algorithm', None)
            hv = filedata.get('hash-value', None)
            if f==filename and hash_algorithm==h and hv==hash_value:
                found = True
                break

        if not found:
            self[flags.EXTERNAL_FILE_FLAG].append({
                'filename':os.path.basename(filepath),
                'hash-algorithm': hash_algorithm,
                'hash-value':hash_value
            })
            self.updated = True

    
        

def dump(data, filestream):
    if flags.UUID4_FLAG not in data: data[flags.UUID4_FLAG] = str(uuid.uuid4())

    now = datetime.datetime.now()
    if flags.CREATED_ON_FLAG not in data: data[flags.CREATED_ON_FLAG] = now.isoformat(' ') 
    
    if data.updated:
        data[flags.UPDATED_ON_FLAG] = now.isoformat(' ') 

    res = json.dump(data, filestream, indent=4, sort_keys=True)
    data.updated = False
    return res

def dumps(data, sort_keys=False):
    return json.dumps(data, sort_keys=sort_keys)


def load(filestream):
    data = json.load(filestream)
    return scadict(data=data)


# !/usr/bin/python3
# -*- coding: utf-8 -*-

from confapp import conf

from pybpodgui_plugin.models.board.board_uibusy import BoardUIBusy

Board = type(
    'Board',
    tuple(conf.GENERIC_EDITOR_PLUGINS_FINDER.find_class('models.board.Board') + [BoardUIBusy]),
    {}
)



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging
import os
import shutil
import pybpodgui_api
from pybpodgui_api.models.setup.setup_base import SetupBase

from sca.formats import json

logger = logging.getLogger(__name__)


class SetupBaseIO(SetupBase):

    def __init__(self, experiment):
        super(SetupBaseIO, self).__init__(experiment)

        # initial name. Used to track if the name was updated
        self.initial_name = None
        self.data = None

    ##########################################################################
    ####### FUNCTIONS ########################################################
    ##########################################################################

    def collect_data(self, data):
        data.update({'name': self.name})
        data.update({'board_uuid4': str(self.board.uuid4 if self.board else None)})
        data.update({'task': str(self.task.uuid4 if self.task else None)})
        self.board_task.collect_data(data)

        data.update({'sessions': []})

        for session in self.sessions:
            data['sessions'].append(session.collect_data({}))

        return data

    def save(self):
        """
        Save setup data on filesystem.

        :ivar str parent_path: Experiment path.  
        :return: Dictionary containing the setup info to save.  
        :rtype: dict
        """
        if not self.name:
            logger.warning("Skipping setup without name")
        else:

            if self.initial_name is not None:
                initial_path = os.path.join(self.experiment.path, 'setups', self.initial_name)

                if initial_path != self.path:
                    shutil.move(initial_path, self.path)
                    current_filepath = os.path.join(self.path, self.initial_name+'.json')
                    future_filepath = os.path.join(self.path, self.name+'.json')
                    shutil.move(current_filepath, future_filepath)

            if not os.path.exists(self.path):
                os.makedirs(self.path)

            # save sessions
            for session in self.sessions:
                session.save()

            self.project.remove_non_existing_repositories(
                os.path.join(self.path, 'sessions'),
                [session.name for session in self.sessions]
            )

            if self.data:
                data = self.data
            else:
                data = json.scadict(
                    uuid4_id=self.uuid4,
                    software='PyBpod GUI API v'+str(pybpodgui_api.__version__),
                    def_url='http://pybpod.readthedocs.org',
                    def_text='This file contains information about a PyBpod experiment setup.'
                )
            data['board'] = self.board.name if self.board else None
            data['task'] = self.task.name if self.task else None
            data['subjects'] = [subject.name for subject in self.subjects]
            data['detached'] = self.detached

            # collect board_task data
            for key, value in self.board_task.save().items():
                data[key] = value

            if self.board:
                data.add_external_ref(self.board.uuid4)

            for subject in self.subjects:
                data.add_external_ref(subject.uuid4)

            config_path = os.path.join(self.path, self.name+'.json')
            with open(config_path, 'w') as fstream:
                json.dump(data, fstream)

            self.initial_name = self.name

    def load(self, path):
        """
        Load setup data from filesystem

        :ivar str setup_path: Path of the setup
        :ivar dict data: data object that contains all setup info
        """
        self.name = os.path.basename(path)
        with open(os.path.join(self.path, self.name+'.json'), 'r') as stream:
            self.data = data = json.load(stream)
        self.uuid4 = data.uuid4 if data.uuid4 else self.uuid4

        self.initial_name = self.name
        self.board = data.get('board', None)
        self.task = data.get('task', None)

        self.detached = data.get('detached', False)
        self.board_task.load(data)

        for subject_name in data.get('subjects', []):
            self += self.project.find_subject(subject_name)

        sessionspath = os.path.join(self.path, 'sessions')
        if os.path.exists(sessionspath):
            for name in sorted(os.listdir(sessionspath)):
                if os.path.isfile(os.path.join(sessionspath, name)):
                    continue
                session = self.create_session()
                session.load(os.path.join(sessionspath, name))

        self._sessions = sorted(self.sessions, key=lambda x: x.started, reverse=True)



# !/usr/bin/python
# -*- coding: utf-8 -*-

import logging

from confapp import conf

from AnyQt.QtGui import QIcon
from pyforms.controls import ControlTree

from pybpodgui_plugin_trial_timeline.trial_timeline import TrialTimeline

logger = logging.getLogger(__name__)

class SubjectTreeNode(object):

    def create_sessiontreenode(self, session):
        node = super(SubjectTreeNode, self).create_sessiontreenode(session)

        self.trialtimeline_action = self.tree.add_popup_menu_option(
            'Trial Timeline', 
            session.open_trialtimeline_window,
            item=node
            )

        self.trialtimeline_detached_action = self.tree.add_popup_menu_option(
            'Trial Timeline (Detached)', 
            session.open_trialtimeline_window_detached,
            item=node
            )

        return node


from pybpod_alyx_module.alyxapi.subjects.get import Get
from pybpod_alyx_module.alyxapi.subjects.post import Post


class Subjects:
    
    def __init__(self, _apibase):
        # print('Subject base init')
        self.apibase = _apibase
        self.get = Get(_apibase)
        self.post = Post(_apibase)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
from datetime import datetime as datetime_now
from pybpodapi.utils import date_parser


class BaseMessage(object):
    """
    Represents a session message
    It may have been originated from the board of from pc
    """

    MESSAGE_TYPE_ALIAS = "MESSAGE"
    MESSAGE_COLOR = (200, 200, 200)

    def __init__(self, content, host_timestamp=None):
        self.pc_timestamp = datetime_now.now()
        self.host_timestamp = host_timestamp
        self.content = content

    def __str__(self):
        return "host-time:{0} pc-time:{1} {2}".format(
            self.host_timestamp if self.host_timestamp is not None else "",
            self.pc_timestamp.strftime("%Y%m%d%H%M%S") if self.pc_timestamp else "",
            self.content,
        )

    @classmethod
    def check_type(cls, typestr):
        """
        Returns True if the typestr represents the class
        """
        return typestr == cls.MESSAGE_TYPE_ALIAS

    def tolist(self):
        return [
            self.MESSAGE_TYPE_ALIAS,
            self.pc_timestamp,
            self.host_timestamp,
            None,  # reserved to store the host end timestamp
            self.content,
            None,
        ]

    @classmethod
    def fromlist(cls, row):
        """
        Returns True if the typestr represents the class
        """
        obj = cls(row[4], float(row[2]) if row[2] else None)
        obj.pc_timestamp = date_parser.parse(row[1])

        return obj



# !/usr/bin/python
# -*- coding: utf-8 -*-

import os
import logging
import hashlib
import shutil

import pybpodgui_api
from sca.formats import json

from pybpodgui_api.models.project.project_base import ProjectBase

logger = logging.getLogger(__name__)


class ProjectIO(ProjectBase):

    def __init__(self):
        super(ProjectIO, self).__init__()

        self.data_hash = None
        self.data = None

    ##########################################################################
    ####### FUNCTIONS ########################################################
    ##########################################################################

    def load(self, project_path):
        """
        Load project from a folder.

        :ivar str project_path: Full path of the project to load.
        """
        self.name = os.path.basename(project_path)
        self.path = project_path

        with open(os.path.join(self.path, self.name+'.json'), 'r') as stream:
            self.data = data = json.load(stream)

        self.uuid4 = data.uuid4 if data.uuid4 else self.uuid4

        logger.debug('=== LOAD USERS ===')
        userspath = os.path.join(self.path, 'users')
        if os.path.exists(userspath):
            for name in os.listdir(userspath):
                if os.path.isfile(os.path.join(userspath, name)):
                    continue
                user = self.create_user()
                user.load(os.path.join(userspath, name))

        logger.debug("==== LOAD TASKS ====")

        # load tasks
        taskspath = os.path.join(self.path, 'tasks')
        if os.path.exists(taskspath):
            for name in os.listdir(taskspath):
                if os.path.isfile(os.path.join(taskspath, name)):
                    continue
                task = self.create_task()
                task.load(os.path.join(taskspath, name))

        logger.debug("==== LOAD BOARDS ====")

        # load boards
        boardspath = os.path.join(self.path, 'boards')
        if os.path.exists(boardspath):
            for name in os.listdir(boardspath):
                if os.path.isfile(os.path.join(boardspath, name)):
                    continue
                board = self.create_board()
                board.load(os.path.join(boardspath, name))

        logger.debug("==== LOAD SUBJECTS ====")

        # load subjects
        subjectspath = os.path.join(self.path, 'subjects')
        if os.path.exists(subjectspath):
            for name in os.listdir(subjectspath):
                if os.path.isfile(os.path.join(subjectspath, name)):
                    continue
                subject = self.create_subject()
                subject.load(os.path.join(subjectspath, name))

        logger.debug("==== LOAD EXPERIMENTS ====")

        # load experiments
        experimentspath = os.path.join(self.path, 'experiments')
        if os.path.exists(experimentspath):
            for name in os.listdir(experimentspath):
                if os.path.isfile(os.path.join(experimentspath, name)):
                    continue
                experiment = self.create_experiment()
                experiment.load(os.path.join(experimentspath, name))

        logger.debug("==== POSTLOAD SUBJECTS ====")
        for subject in self.subjects:
            subject.post_load()

        self.data_hash = self.__generate_project_hash()

        logger.debug("==== LOAD FINNISHED ====")

    def save(self, project_path):
        """
        Save project data on file
        :param str project_path: path to project
        :return: project data saved on settings file
        """

        if project_path is not None:
            self.path = project_path

        logger.debug("saving project path: %s",  project_path)
        logger.debug("current project name: %s", self.name)
        logger.debug("current project path: %s", self.path)

        ########### SAVE THE USERS ############
        userspath = os.path.join(self.path, 'users')
        if not os.path.exists(userspath):
            os.makedirs(userspath)
        for user in self.users:
            user.save()
        self.remove_non_existing_repositories(userspath, [user.name for user in self.users])

        ########### SAVE THE TASKS ############
        taskspath = os.path.join(self.path, 'tasks')
        if not os.path.exists(taskspath):
            os.makedirs(taskspath)
        for task in self.tasks:
            task.save()
        self.remove_non_existing_repositories(taskspath, [task.name for task in self.tasks])

        ########### SAVE THE BOARDS ###########
        boardspath = os.path.join(self.path, 'boards')
        if not os.path.exists(boardspath):
            os.makedirs(boardspath)
        for board in self.boards:
            board.save()
        self.remove_non_existing_repositories(boardspath, [board.name for board in self.boards])

        ########### SAVE THE SUBJECTS ###############
        subjectspath = os.path.join(self.path, 'subjects')
        if not os.path.exists(subjectspath):
            os.makedirs(subjectspath)
        for subject in self.subjects:
            subject.save()
        self.remove_non_existing_repositories(subjectspath, [subject.name for subject in self.subjects])

        ########### SAVE THE EXPERIMENTS ############
        experimentspath = os.path.join(self.path, 'experiments')
        if not os.path.exists(experimentspath):
            os.makedirs(experimentspath)
        for experiment in self.experiments:
            experiment.save()
        self.remove_non_existing_repositories(experimentspath, [experiment.name for experiment in self.experiments])

        ########### SAVE THE PROJECT ############

        if self.data:
            data = self.data
        else:
            data = json.scadict(
                uuid4_id=self.uuid4,
                software='PyBpod GUI API v'+str(pybpodgui_api.__version__),
                def_url='http://pybpod.readthedocs.org',
                def_text='This file contains information about a PyBpod project.'
            )
        data['name'] = self.name

        name = os.path.basename(self.path)
        config_path = os.path.join(self.path, name+'.json')
        with open(config_path, 'w') as fstream:
            json.dump(data, fstream)

        self.data_hash = self.__generate_project_hash()

    def remove_non_existing_repositories(self, path, names):
        try:
            nodes = os.listdir(path)
        except:
            nodes = []

        for nodename in nodes:
            if nodename not in names:
                nodepath = os.path.join(path, nodename)
                if not os.path.isfile(nodepath):
                    shutil.rmtree(nodepath)

    def is_saved(self):
        """
        Verifies if project has changes by doing a recursive checksum on all entities

        :rtype: bool
        """
        if not self.path:
            return False

        current_hash = self.__generate_project_hash()

        if self.data_hash != current_hash:
            logger.warning("Different project data hashes:\n%s\n%s", self.data_hash, current_hash)
            return False

        return True

    def collect_data(self, data):
        """
        Collect the data of the project. This function is used to calculate the checksum of the project and verify if it was updated.

        :rtype: dict
        """
        data.update({'name': self.name})
        data.update({'experiments': []})
        data.update({'boards': []})
        data.update({'users': []})
        data.update({'subjects': []})
        data.update({'tasks': []})

        for board in self.boards:
            data['boards'].append(board.collect_data({}))

        for experiment in self.experiments:
            data['experiments'].append(experiment.collect_data({}))

        for task in self.tasks:
            data['tasks'].append(task.collect_data({}))

        for user in self.users:
            data['users'].append(user.collect_data({}))

        for subject in self.subjects:
            data['subjects'].append(subject.collect_data({}))

        logger.debug("Project data: %s", data)

        return data

    def __save_project_hash(self):
        self.data_hash = self.__generate_project_hash()
        logger.debug("Project data hash: %s", self.data_hash)

    def __generate_project_hash(self):
        return hashlib.sha256(
            json.dumps(self.collect_data(data={}), sort_keys=True).encode('utf-8')).hexdigest()



# !/usr/bin/python3
# -*- coding: utf-8 -*-

from pybpodgui_api.exceptions.api_error import APIError


class PycontrolError(APIError):
    """ Exception raised when a board operation fails"""
    pass



# !/usr/bin/python3
# -*- coding: utf-8 -*-

""" session_window.py

"""

import logging

import numpy as np
from AnyQt.QtCore import (QAbstractTableModel, QSize, Qt, QTimer,
                          pyqtSignal)
from AnyQt.QtGui import QBrush, QColor
from PyQt5.QtWidgets import QHeaderView
from confapp import conf
from PyQt5.uic.Compiler.qtproxies import QtWidgets

from pybpodapi.session import Session
from pyforms.basewidget import BaseWidget
from pyforms.controls import ControlCheckBox, ControlTableView

logger = logging.getLogger(__name__)


class PandasModel(QAbstractTableModel):
    """
    Class to populate a table view with a pandas dataframe
    """
    COLOR_MSG                = QBrush(QColor(200, 200, 200))
    COLOR_DEBUG              = QBrush(QColor(200, 200, 200))
    COLOR_TRIAL              = QBrush(QColor(0, 100, 200))
    COLOR_ERROR              = QBrush(QColor(240, 0, 0))
    COLOR_INFO               = QBrush(QColor(150, 150, 255))
    COLOR_SOFTCODE_OCCURENCE = QBrush(QColor(40, 30, 30))
    COLOR_STATE_OCCURENCE    = QBrush(QColor(70, 180, 70))
    COLOR_STATE_TRANSITION   = QBrush(QColor(0, 100, 0))
    COLOR_STDERR             = QBrush(QColor(255, 0, 0))
    COLOR_STDOUT             = QBrush(QColor(150, 150, 150))
    COLOR_TRIAL              = QBrush(QColor(0, 0, 255))
    COLOR_WARNING            = QBrush(QColor(255, 100, 0))

    COLUMNS_WIDTHS = [QSize(100, 30), QSize(400, 30), QSize(200, 30), QSize(200, 30), QSize(200, 30), QSize(200, 30)]

    def __init__(self, data, parent=None):
        QAbstractTableModel.__init__(self, parent)
        self._data = np.array(data.values)
        self._cols = data.columns
        self.r, self.c = np.shape(self._data)

    def rowCount(self, parent=None):
        return self.r

    def columnCount(self, parent=None):
        return self.c

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            if role == Qt.ForegroundRole:
                dtype = self._data[index.row()][0]

                if   dtype == Session.MSGTYPE_DEBUG:      return self.COLOR_DEBUG
                elif dtype == Session.MSGTYPE_ENDTRIAL:   return self.COLOR_TRIAL
                elif dtype == Session.MSGTYPE_ERROR:      return self.COLOR_ERROR
                elif dtype == Session.MSGTYPE_INFO:       return self.COLOR_INFO
                elif dtype == Session.MSGTYPE_SOFTCODE:   return self.COLOR_SOFTCODE_OCCURENCE
                elif dtype == Session.MSGTYPE_STATE:      return self.COLOR_STATE_OCCURENCE
                elif dtype == Session.MSGTYPE_TRANSITION: return self.COLOR_STATE_TRANSITION
                elif dtype == Session.MSGTYPE_STDERR:     return self.COLOR_STDERR
                elif dtype == Session.MSGTYPE_STDOUT:     return self.COLOR_STDOUT
                elif dtype == Session.MSGTYPE_TRIAL:      return self.COLOR_TRIAL
                elif dtype == Session.MSGTYPE_WARNING:    return self.COLOR_WARNING
                else:                                   return self.COLOR_MSG

            elif role == Qt.DisplayRole:
                return str(self._data[index.row(), index.column()])

        return None

    def headerData(self, col, orientation, role):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._cols[col]
        #elif orientation == Qt.Horizontal and role == Qt.SizeHintRole:
        #    return QVariant(self.COLUMNS_WIDTHS[col])

    def flags(self, index):
        return Qt.ItemIsEnabled


class SessionHistory(BaseWidget):
    """ Plugin main window """

    def __init__(self, session):
        BaseWidget.__init__(self, session.name)
        self.set_margin(5)

        self._autoscroll  = ControlCheckBox('Auto-scroll', default=True)
        self._log         = ControlTableView(select_entire_row=True)

        self._formset = [
            ('_autoscroll',' ',' ',' ',),
            '_log'
        ]

        self._counter = 0
        self.session    = session
        self.model = PandasModel(session.data)
        self._log.value = self.model
        
        self._timer = QTimer()
        self._timer.timeout.connect(self.__update_table_view)

    def __update_table_view(self):
        # check if the session is running. If not stops the timer
        if not self.session.is_running: self._timer.stop()

        self._log.value = None
        self._log.value = self.model
        if self._autoscroll.value: self._log.scrollToBottom()

    
    def show(self, detached=False):
        if self.session.is_running and self.session.setup.detached:
            return

        # Prevent the call to be recursive because of the mdi_area
        if not detached:
            if hasattr(self, '_show_called'):
                BaseWidget.show(self)
                return
            self._show_called = True
            self.mainwindow.mdi_area += self
            del self._show_called
        else:
            BaseWidget.show(self)

        
        self._stop = False # flag used to close the gui in the middle of a loading
        if not self._stop and self.session.is_running:
            self._timer.start(conf.SESSIONLOG_PLUGIN_REFRESH_RATE)

    def hide(self):
        self._timer.stop()
        self._stop = True
        
    def before_close_event(self):       
        self._timer.stop()
        self._stop = True
        self.session.sessionhistory_action.setEnabled(True)
        self.session.sessionhistory_detached_action.setEnabled(True)

    @property
    def mainwindow(self): return self.session.mainwindow

    @property
    def title(self): return BaseWidget.title.fget(self)
    @title.setter
    def title(self, value):  BaseWidget.title.fset(self, 'Session History: {0}'.format(value))



import math
import numpy as np


class WindowConfiguration:
    def __init__(self,
                 left_duration=0.1,
                 left_apply_window_start=True,
                 left_apply_window_end=True,
                 left_window_function='Hanning',
                 right_duration=0.1,
                 right_apply_window_start=True,
                 right_apply_window_end=True,
                 right_window_function='Hanning'):
        """

        :param left_duration: (Optional) Duration of the window in seconds, for the left channel. If zero, no window
            will be created for this channel. Default: 0.1s
        :param left_apply_window_start: (Optional) True if the window should be applied to the **start** of the sound
            for the left channel, False otherwise. Default: True
        :param left_apply_window_end: (Optional) True if the window should be applied to the **end** of the sound for
            the left channel, False otherwise. Default: True
        :param left_window_function: (Optional) Window function that should be used for the left channel. Possible values
            accepted: 'Hanning', 'Hamming', 'Blackman', 'Bartlett'. Default: 'Hanning"
        :param right_duration: (Optional) Duration of the window in seconds, for the right channel. If zero, no window
            will be created for this channel. Default: 0.1s
        :param right_apply_window_start: (Optional) True if the
            window should be applied to the **start** of the sound, for the right channel, False otherwise. Default: True
        :param right_apply_window_end: (Optional) True if the window
            should be applied to the **end** of the sound for the right channel, False otherwise. Default: True
        :param right_window_function: (Optional) Window function that should be used for the left channel.
            Possible values accepted: 'Hanning', 'Hamming', 'Blackman', 'Bartlett'. Default: 'Hanning"
        """
        self.left_duration = left_duration
        self.right_duration = right_duration
        self.left_apply_window_start = left_apply_window_start
        self.right_apply_window_start = right_apply_window_start
        self.left_apply_window_end = left_apply_window_end
        self.right_apply_window_end = right_apply_window_end
        self.left_window_function = left_window_function
        self.right_window_function = right_window_function


def generate_sound(filename=None, fs=96000, duration=1, frequency_left=1000, frequency_right=1000, window_configuration:WindowConfiguration=None):
    """
    Helper method to dynamically generated a sound that can be used in with the Sound Card module.

    :param filename: (Optional)
    :param fs: (Optional) number of samples per second (standard)
    :param duration: (Optional) sound duration in seconds
    :param frequency_left: (Optional) number of cycles per second (Hz) (frequency of the sine wave for the left channel)
    :param frequency_right: (Optional) number of cycles per second (Hz) (frequency of the sine wave for the right channel)
    :param window_configuration: (Optional) WindowConfiguration object to apply to the generated sound.

    :return: Returns the **flatten** generated sound as a numpy array (as np.int8)

    """

    amplitude24bits = math.pow(2, 31) - 1

    samples = np.arange(0, duration, 1 / fs)
    wave_left = amplitude24bits * np.sin(2 * math.pi * frequency_left * samples)
    wave_right = amplitude24bits * np.sin(2 * math.pi * frequency_right * samples)

    if window_configuration:
        wave_left = generate_window(fs,
                                    wave_left,
                                    window_configuration.left_duration,
                                    window_configuration.left_apply_window_start,
                                    window_configuration.left_apply_window_end,
                                    window_configuration.left_window_function)
        wave_right = generate_window(fs,
                                     wave_right,
                                     window_configuration.right_duration,
                                     window_configuration.right_apply_window_start,
                                     window_configuration.right_apply_window_end,
                                     window_configuration.right_window_function)

    stereo = np.stack((wave_left, wave_right), axis=1)

    wave_int = stereo.astype(np.int32)

    if filename:
        with open(filename, 'wb') as f:
            wave_int.tofile(f)

    return wave_int.flatten()


def generate_window(fs, wave_int, duration, apply_start, apply_end, window_function):
    """

    :param fs: number of samples per second (standard)
    :param wave_int: base sound where the window will be applied
    :param duration: duration of the window (it will be the same on the start and end)
    :param apply_start: True if the window should be created at the start, False otherwise.
    :param apply_end: True if the window should be created at the end, False otherwise.
    :param window_function: window function to be generated. Possible values accepted:
        'Hanning', 'Hamming', 'Blackman', 'Bartlett'. It will revert to 'Hanning' if an unknown option is given.

    :return: Returns the modified sound with the window applied to it.
    """
    len_fade = int(duration * fs)
    if window_function == 'Hanning':
        fade_io = np.hanning(len_fade * 2)
    elif window_function == 'Hamming':
        fade_io = np.hamming(len_fade * 2)
    elif window_function == 'Blackman':
        fade_io = np.blackman(len_fade * 2)
    elif window_function == 'Bartlett':
        fade_io = np.bartlett(len_fade * 2)
    else:   # default
        fade_io = np.hanning(len_fade * 2)

    fadein = fade_io[:len_fade]
    fadeout = fade_io[len_fade:]
    win = np.ones(len(wave_int))
    if apply_start:
        win[:len_fade] = fadein
    if apply_end:
        win[-len_fade:] = fadeout
    wave_int = wave_int * win
    return wave_int



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging

from confapp import conf

from AnyQt.QtGui import QIcon

from pybpodgui_api.models.setup import Setup
from pybpodgui_plugin.models.setup.setup_com import SetupCom

logger = logging.getLogger(__name__)


class SetupUIBusy(SetupCom):
    """
    Extends setup with UI refreshment logic.

    .. seealso::
        UI buttons: :py:class:`pybpodgui_plugin.models.setup.setup_window.SetupWindow`.

    **Methods**

    """

    def stop_trial(self):
        super(SetupUIBusy, self).stop_trial()
        self._pause_btn.label = 'Pause'
        self._pause_btn.checked = False

    def pause_trial(self):
        super(SetupUIBusy, self).pause_trial()
        self._pause_btn.label = 'Resume'

    def resume_trial(self):
        super(SetupUIBusy, self).resume_trial()
        self._pause_btn.label = 'Pause'

    def stop_task(self):
        super(SetupUIBusy, self).stop_task()
        self._run_task_btn.enabled = False

    def kill_task(self):
        super(SetupUIBusy, self).kill_task()
        self._kill_task_btn.enabled = False

    def update_ui(self):
        """
        Update UI button states and labels and tree icons when board communication is active.
        """
        if not hasattr(self, 'node'):
            return

        logger.debug('Setup [{0}] status: {1}'.format(self.name, self.status))

        if self.status == Setup.STATUS_READY:

            self.node.setIcon(0, QIcon(conf.BOX_SMALL_ICON))

            self._run_task_btn.label = 'Run'
            self._run_task_btn.checked = False
            self._run_task_btn.enabled = True

            self._kill_task_btn.enabled = False

            self._stoptrial_btn.enabled = False
            self._pause_btn.label = 'Pause'
            self._pause_btn.enabled = False
            self._pause_btn.checked = False

            self._board.enabled = True
            self._detached.enabled = True
            self._task.enabled = True

        elif self.status == Setup.STATUS_BOARD_LOCKED:

            self.node.setIcon(0, QIcon(conf.BUSY_SMALL_ICON))
            # self.disable_all_task_buttons()
            self._board.enabled = False
            self._task.enabled = False

        elif self.status == Setup.STATUS_RUNNING_TASK:

            self._run_task_btn.label = 'Stop'
            self._run_task_btn.checked = True

            self._kill_task_btn.enabled = True

            self._stoptrial_btn.enabled = True
            self._pause_btn.enabled = True

            self.node.setIcon(0, QIcon(conf.PLAY_SMALL_ICON))
            # self.disable_all_task_buttons()
            self._board.enabled = False
            self._task.enabled = False
            self._detached.enabled = False

        self.board_task.update_ui()
        if self.last_session:
            self.last_session.update_ui()

    def enable_all_task_buttons(self):
        """
        Enable UI buttons for upload task, configure task and run task
        """
        self._run_task_btn.enabled = True

    def disable_all_task_buttons(self):
        """
        Disable UI buttons for upload task, configure task and run task
        """
        self._run_task_btn.enabled = False



from enum import IntEnum
from pybpodapi.bpod_modules.bpod_module import BpodModule


class SoundCommandType(IntEnum):
    """
    Enumeration for the commands that can be sent through the Sound Card module, when connected through the BPod's
    State Machine.

    """
    #: Plays a specific sound index
    PLAY = 1
    #: Stops playing a specific sound index
    STOP_SPECIFIC = 2
    #: Stops all sounds
    STOP_ALL = 3


class SoundCard(BpodModule):

    @staticmethod
    def check_module_type(module_name):
        return module_name and module_name.startswith('SoundCard')

    @staticmethod
    def get_command(command_type, sound_index=None):
        """
        Returns the proper bytes to send as output_actions in the BPod StateMachine's states. In the case of the Play
        and StopSpecific command_types, it is required to \ use the BPod's load_serial_message method to be able to
        send to the module more than 1 byte properly.

        .. note:: This might not be required in a future version of BPod's firmware

        :param command_type: Instruction of type :class:`.SoundCommandType` to generate the command
        :param sound_index: The sound index to play or stop

        """
        if not command_type:
            raise Exception("You need to provide the type of the command you want returned")

        if command_type is SoundCommandType.PLAY or command_type is SoundCommandType.STOP_SPECIFIC:
            if not sound_index:
                raise Exception("You need to provide the sound_index value to play")

            if sound_index < 2 or sound_index > 31:
                raise Exception("sound_index must have a value between 2 and 31")

        if command_type is SoundCommandType.PLAY:
            return [ord('P'), sound_index]
        if command_type is SoundCommandType.STOP_SPECIFIC:
            return [ord('x'), sound_index]

        # by default return stop
        return ord('X')



# !/usr/bin/python3
# -*- coding: utf-8 -*-

__version__ = "1.0.1"
__author__ = "Carlos Mão de Ferro"
__credits__ = ["Carlos Mão de Ferro", "Ricardo Ribeiro"]
__license__ = "Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>"
__maintainer__ = ["Carlos Mão de Ferro", "Ricardo Ribeiro"]
__email__ = ["cajomferro@gmail.com", "ricardojvr@gmail.com"]
__status__ = "Development"

import loggingbootstrap
from confapp import conf

conf += 'pybpodgui_plugin_timeline.settings'
conf += 'pybpodgui_plugin_timeline.resources'

# setup different loggers but output to single file
loggingbootstrap.create_double_logger("pybpodgui_plugin_timeline", conf.APP_LOG_HANDLER_CONSOLE_LEVEL,
                                      conf.APP_LOG_FILENAME,
                                      conf.APP_LOG_HANDLER_FILE_LEVEL)

from pybpodgui_plugin_timeline.trials_plot_window import TrialsPlotWindow as TrialsPlot



import logging
from pybpodgui_plugin.models.user.user_treenode import UserTreeNode

logger = logging.getLogger(__name__)


class UserDockWindow(UserTreeNode):

    def show(self):
        self.mainWindow.details.value = self
        super(UserDockWindow, self).show()

    def focus_name(self):
        self._namebox.form.lineEdit.setFocus()

    def remove(self):
        """

        Prompts user to confirm user removal and closes mdi windows associated with this user.

        .. seealso::
            This method extends user tree node :py:meth:`pybpodgui_plugin.models.user.user_treenode.UserTreeNode.remove`.

        """
        reply = self.question('User "{0}" will be deleted. Are you sure?'.format(self.name), 'Warning')
        if reply == 'yes':
            if hasattr(self, '_code_editor'):
                self.mainwindow.mdi_area -= self._code_editor
            super(UserDockWindow, self).remove()

    def close(self):
        self.mainWindow.details.value = None
        super(UserDockWindow, self).close()

    @property
    def mainWindow(self):
        return self.project.mainwindow



__version__ = '0.1.6'
__author__ 		= ['Luís Teixeira']
__credits__ 	= ["Luís Teixeira"]
__license__ 	= "MIT"
__maintainer__ 	= ['Luís Teixeira']
__email__ 		= ['micboucinha@gmail.com']
__status__ 		= "Development"

from confapp import conf

conf += 'pybpod_soundcard_module.settings'

from pybpod_soundcard_module.module import SoundCard as BpodModule



from pyforms.basewidget import BaseWidget
from pyforms.controls   import ControlLabel
from pyforms.controls   import ControlText
from pyforms.controls   import ControlList
from pyforms.controls   import ControlButton
from pyforms.controls   import ControlProgress
from pyforms.controls   import ControlTextArea
from pyforms.controls   import ControlCombo
from pyforms.controls   import ControlWeb

from AnyQt.QtWidgets import QApplication
from AnyQt.QtCore    import QTimer

from .package_win    import PackageWindow

from confapp import conf

from importlib import reload

import yaml
import sys
import subprocess
import xmlrpc.client
import pkg_resources
import datetime

import markdown2

conf.PYBPOD_REPOSITORIES_TXT_LIST = '../../repositories.yml'


class UpdateCenter(BaseWidget):

    MAIN_PACKAGE_NAME = 'pyforms-gui'

    pypi = xmlrpc.client.ServerProxy('https://pypi.org')

    def __init__(self, *args, **kwargs):
        super().__init__('Update center')

        self.set_margin(10)

        self._versions      = ControlCombo('Versions', changed_event=self.__versions_changed_evt)
        self._web           = ControlWeb('Web', default='https://google.com')
        self._install       = ControlButton('Install')

        #Definition of the forms fields
        self._packageslst = ControlList('PACKAGES', readonly=True, select_entire_row=True)
        self._searchbox   = ControlText('Search')
        self._searchbtn   = ControlButton('Load packages', default=self.__searchbtn_evt)
        self._progress    = ControlProgress('Progress')

        self._formset = [
            {
                'Available versions': [
                    (' ', '_versions','_install'),
                    '_web',
                ],
                'Packages': [
                    '_packageslst'
                ]
            },
            '_progress'
        ]
        QTimer.singleShot(300, self.load_packages_list)

    def __searchbtn_evt(self):
        self.load_packages_list()


    def __versions_changed_evt(self):
        """
        Called to update the version documentation
        """
        version = self._versions.value
        data = self.pypi.release_data(self.MAIN_PACKAGE_NAME, version)
        self._web.html = markdown2.markdown( data.get('description', '') )


    def load_application_info(self):

        name = 'pyforms-gui'

        new_version = self.pypi.package_releases(name)
        if not new_version:
            new_version = self.pypi.package_releases(name.capitalize())

        if new_version is None: return new_version

        new_version  = new_version[0]
        all_versions = self.pypi.package_releases(name, True)
        data = self.pypi.release_data(name, new_version)


        return new_version, all_versions, data.get('description', '')


    def load_packages_list(self):
        reload(pkg_resources)
        self._packageslst.clear()
        QApplication.processEvents()

        new_version, all_versions, description = self.load_application_info()
        
        self._web.html = markdown2.markdown(description, extras=["footnotes"])


        with open(conf.PYBPOD_REPOSITORIES_TXT_LIST) as infile:

            data = yaml.load(infile)
            
            self._versions.clear()
            for recipe in data.get('recipes', []):
                for name, d in recipe.items():
                    self._versions.add_item(name)
                    #self._web.html = d.get('description', '')


            packages_lst = data.get('packages',[])

            self._progress.value = 0
            self._progress.max   = len(packages_lst)

            for index, name in enumerate(packages_lst):
                name = name.strip(' \t\n\r')

                QApplication.processEvents()

                
                self._packageslst += [PackageWindow(title=name, parent_widget=self)]

                self._progress.value = index+1
                QApplication.processEvents()

            self._progress.hide()





import pyforms
from AnyQt import QtGui, QtCore
from AnyQt.QtWidgets import QLineEdit
from confapp import conf
from pybpod_alyx_module.models.subject.alyx_subject import AlyxSubject
from pybpod_alyx_module.module_api import AlyxModule
from pyforms.basewidget import BaseWidget
from pyforms.controls import ControlText, ControlButton, ControlLabel


class AlyxModuleGUI(AlyxModule, BaseWidget):

    TITLE = 'Alyx connection'

    def __init__(self, _project=None):
        BaseWidget.__init__(self, self.TITLE, parent_win=_project)
        AlyxModule.__init__(self)

        self.project = _project
        
        self._addressbox = ControlText('Address')
        self._username = ControlText('User:')
        self._password = ControlText('Password:')
        self._username = ControlText('User:',default = conf.ALYX_PLUGIN_USERNAME)
        self._password = ControlText('Password:', default = conf.ALYX_PLUGIN_PASSWORD)
        self._connect_btn = ControlButton('Connect',default = self._connect)
        self._status_lbl = ControlLabel('Status: Not Connected')
        self._getsubjects_btn = ControlButton('Get Subjects', default = self._get_subjects)
        self._getsubjects_btn.enabled = False
        self.set_margin(10)

        self._addressbox.value = conf.ALYX_PLUGIN_ADDRESS
        self._addressbox.changed_event = self.setaddr

        if self.project.loggeduser is not None:
            self._username.value = self.project.loggeduser.name
        
        self._password.form.lineEdit.setEchoMode(QLineEdit.Password)

        self.formset = [
            '_addressbox',
            '_username',
            '_password',
            '_connect_btn',
            '_status_lbl',
            '_getsubjects_btn'
        ]

    def setaddr(self):
        self.set_alyx_address(self._addressbox.value)

    def _connect(self):
        if self._connect_to_alyx(self._username.value,self._password.value):
            usersearch = self.project.find_user(self._username.value)
            if usersearch is None: # create this user on the project
                newuser = self.project.create_user()
                newuser.name = self._username.value
                newuser.node_double_clicked_event()
            else:
                usersearch.node_double_clicked_event()
            self.project.loggeduser.connection = 'ALYX'
            self._status_lbl.value = 'Status: CONNECTED'
            self.project.loggeduser = self.project.loggeduser
            self._getsubjects_btn.enabled = True

    def _get_subjects(self):
        result = self.get_alyx_subjects(self._username.value)
        subjects = self.project.subjects.copy()
        replace_all = False
        ignore_all = False
        for subj in result:
            subjname = subj['nickname']
            is_alive = subj['alive']
            existing = False
            for s in subjects:
                if s.name == subjname:
                    existing = True
                    if replace_all:
                        self.__add_or_remove_subject(is_alive, s, subjname)
                        continue
                    reply = self.question("Subject '{name}' already exists locally. Replace local details?".format(name=s.name),
                                          'Update Subject', buttons=['no', 'yes', 'no_all', 'yes_all'])
                    if reply == 'yes':
                        self.__add_or_remove_subject(is_alive, s, subjname)
                    if reply == 'yes_all':
                        self.__add_or_remove_subject(is_alive, s, subjname)
                        replace_all = True
                    if reply == 'no_all':
                        ignore_all = True
                        break
            if ignore_all:
                break
            if not existing:
                if is_alive:
                    subj_info = self.get_alyx_subject_info(subjname)
                    # SubjectBase constructor adds Subject automatically to self.project so there's no need to add it here
                    newsubject = AlyxSubject(self.project)
                    newsubject.add_alyx_info(subj_info)

    def __add_or_remove_subject(self, is_alive, s, subjname):
        if is_alive:
            subj_info = self.get_alyx_subject_info(subjname)
            s.add_alyx_info(subj_info)
        else:
            s.remove(True)

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        if event.key() == QtCore.Qt.Key_Escape:
            self.close()


if __name__ == '__main__':
    pyforms.start_app(AlyxModuleGUI, geometry=(0, 0, 300, 300))



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging

from confapp import conf

from AnyQt.QtGui import QIcon
from AnyQt import QtCore

from pybpodgui_plugin.models.setup import Setup
from pybpodgui_plugin.models.experiment.experiment_window import ExperimentWindow

logger = logging.getLogger(__name__)


class ExperimentTreeNode(ExperimentWindow):
    """
    Extends experiment window to show up in the project tree section.
    Define here actions related to the experiment tree node.

    **Properties**

        name
            Handles experiment tree node name.

        tree
            Returns project tree.

    **Methods**

    """

    def __init__(self, project):
        # type: (project) -> None
        """

        :param project: project where this experiment belongs
        :type project: pybpodgui_plugin.models.project.Project
        """
        ExperimentWindow.__init__(self, project)

        self.__running_icon = QIcon(conf.PLAY_SMALL_ICON)

        self.create_treenode(self.tree)

    def create_treenode(self, tree):
        """
        Creates node for this experiment under the parent "Experiments" node.

        This methods is called when the experiment is first created.

        The following actions get assigned to node:
            * *Add setup*: :class:`ExperimentTreeNode._ExperimentTreeNode__add_setup`.
            * *Remove*: :class:`ExperimentTreeNode.remove`.

        Sets key events:
            * :class:`ExperimentTreeNode.node_key_pressed_event`


        :param tree: the project tree
        :type tree: pyforms.controls.ControlTree
        :return: new created node
        :return type: QTreeWidgetItem
        """
        self.node = tree.create_child(self.name, self.project.experiments_node,
                                      icon=QIcon(conf.EXPERIMENT_SMALL_ICON))
        self.node.key_pressed_event = self.node_key_pressed_event
        self.node.window = self
        self.node.setExpanded(True)

        tree.add_popup_menu_option('Add setup', self.__add_setup, item=self.node,
                                   icon=QIcon(conf.ADD_SMALL_ICON))
        tree.add_popup_menu_option('Remove', self.remove, item=self.node, icon=QIcon(conf.REMOVE_SMALL_ICON))

        return self.node

    def __add_setup(self):
        """
        Bind events for adding new setup (a.k.a setup). Invokes :class:`ExperimentTreeNode.create_setup`.

        """
        setup = self.create_setup()
        setup.focus_name()

    def create_setup(self):
        """
        Invoke setup (a.k.a setup) creation and focus GUI on the new node.

        :return: the setup just created.
        :return type: Setup

        .. seealso::
            Setup: :class:`pybpodgui_plugin.models.setup.setup_window.SetupWindow`.

        """
        setup = Setup(self)
        self.tree.setCurrentItem(setup.node)
        return setup

    def remove(self):
        """
        Iterates over all setups (a.k.a. setup) and remove them.
        Finally, removes experiment node from project tree and remove experiment from project.

        .. seealso::
            * Experiment removal (API): :class:`pybpodgui_api.models.experiment.experiment_base.ExperimentBase.remove`.
            * Experiment Dock window: :class:`pybpodgui_plugin.models.experiment.experiment_dockwindow.ExperimentDockWindow`.
            * Remove experiment from project: :class:`pybpodgui_api.models.project.project_base.ProjectBase.__sub__`.
        """
        for index in range(len(self.setups) - 1, -1, -1):
            self.setups[index].remove()

        self.project -= self
        self.project.experiments_node.removeChild(self.node)

    def node_key_pressed_event(self, event):
        """
        Sets key events for:
            * Remove experiment: :class:`ExperimentTreeNode.remove`

        :param event: key event
        """
        if event.key() == QtCore.Qt.Key_Delete:
            self.remove()

    @property
    def name(self):
        if hasattr(self, 'node'):
            return str(self.node.text(0))
        else:
            return ExperimentWindow.name.fget(self)

    @name.setter
    def name(self, value):
        ExperimentWindow.name.fset(self, value)
        if hasattr(self, 'node'): self.node.setText(0, value)

    @property
    def tree(self):
        return self.project.tree



# !/usr/bin/python3
# -*- coding: utf-8 -*-

from confapp import conf
from pyforms_generic_editor.models.projects.projects_dockwindow import ProjectsDockWindow as GenericProjects

Projects = type(
    'Projects',
    tuple(conf.GENERIC_EDITOR_PLUGINS_FINDER.find_class('models.projects.Projects') + [GenericProjects]),
    {}
)



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging

from pybpodapi.bpod.hardware.channels import Channels

logger = logging.getLogger(__name__)


class Hardware(object):
    """
    Represents an hardware description based on information received from the current connected Bpod deviced.
    """

    DEFAULT_FREQUENCY_DIVIDER = 1000000

    def __init__(self):

        self.inputs = None
        self.outputs = None
        self.channels = None

        self.max_states = None
        self.max_serial_events = None

        self.inputs_enabled = None
        self.cycle_period = None

        self.n_global_timers = None
        self.n_global_counters = None
        self.n_conditions = None
        self.n_uart_channels = None

        self.firmware_version = None
        self.machine_type = None

        self.live_timestamps = (
            True  # The state machine will end timestamps during the execution
        )

        self.pos_global_timer_trig = 0

    ##################################################
    #################### METHODS #####################
    ##################################################

    def setup(self, modules):
        """
        Set up hardware based on hardware description obtained from Bpod device

        :param HardwareInfoContainer hw_info_container: hardware parameters received from Bpod
        """

        self.outputs = self.outputs  # + ['G', 'G', 'G']

        self.n_uart_channels = len([idx for idx in self.inputs if idx == "U"])

        # set up channels
        self.channels = Channels()  # type: Channels
        self.channels.setup_input_channels(self, modules)
        self.channels.setup_output_channels(self.outputs, self)

        logger.debug(self.channels)

        logger.debug(str(self))

    def __str__(self):
        return (
            "Hardware Configuration\n"
            "Max states: {max_states}\n"
            "Cycle period: {cycle_period}\n"
            "Cycle frequency: {cycle_frequency}\n"
            "Number of events per serial channel: {max_serial_events}\n"
            "Number of global timers: {n_global_timers}\n"
            "Number of global counters: {n_global_counters}\n"
            "Number of conditions: {n_conditions}\n"
            "Inputs ({n_inputs}): {inputs}\n"
            "Outputs ({n_outputs}): {outputs}\n"
            "Enabled inputs ({n_inputs_enabled}): {inputs_enabled}\n"
            "".format(
                max_states=self.max_states,
                cycle_period=self.cycle_period,
                cycle_frequency=self.cycle_frequency,
                max_serial_events=self.max_serial_events,
                n_global_timers=self.n_global_timers,
                n_global_counters=self.n_global_counters,
                n_conditions=self.n_conditions,
                inputs=self.inputs,
                n_inputs=len(self.inputs),
                outputs=self.outputs,
                n_outputs=len(self.outputs),
                inputs_enabled=self.inputs_enabled,
                n_inputs_enabled=len([idx for idx in self.inputs_enabled if idx == 1]),
            )
        )

    ##################################################
    #################### PROPERTIES ##################
    ##################################################

    @property
    def cycle_frequency(self):
        return int(self.DEFAULT_FREQUENCY_DIVIDER / self.cycle_period)

    @property
    def times_scale_factor(self):
        return float(self.cycle_period) / float(self.DEFAULT_FREQUENCY_DIVIDER)

    @property
    def bnc_inputports_indexes(self):
        return [i for i, input_type in enumerate(self.inputs) if input_type == "B"]

    @property
    def wired_inputports_indexes(self):
        return [i for i, input_type in enumerate(self.inputs) if input_type == "W"]

    @property
    def behavior_inputports_indexes(self):
        return [i for i, input_type in enumerate(self.inputs) if input_type == "P"]

    @property
    def bnc_inputports_names(self):
        return [
            "BNC{0}".format(i)
            for i, input_type in enumerate(self.inputs)
            if input_type == "B"
        ]

    @property
    def wired_inputports_names(self):
        return [
            "Wire{0}".format(i)
            for i, input_type in enumerate(self.inputs)
            if input_type == "W"
        ]

    @property
    def behavior_inputports_names(self):
        return [
            "Port{0}".format(i)
            for i, input_type in enumerate(self.inputs)
            if input_type == "P"
        ]



#!/usr/bin/python3
# -*- coding: utf-8 -*-

""" pycontrolgui_plugin_timeline.trials_plot"""
from pyforms.controls import ControlImage
from pyforms.controls import ControlCombo
from pyforms.basewidget import BaseWidget

from pybpodapi.com.messaging.state_occurrence import StateOccurrence
from pybpodapi.com.messaging.event_occurrence import EventOccurrence
from pybpodapi.com.messaging.event_resume import EventResume
from pybpodapi.com.messaging.session_info import SessionInfo
from pybpodgui_api.exceptions.run_setup import RunSetupError

import logging
import math
import pydot

from datetime import datetime as datetime_now

from confapp import conf

from AnyQt.QtCore import QTimer, QEventLoop

logger = logging.getLogger(__name__)


class StmDiagramWindow(BaseWidget):
    """ Show all boxes live state for an experiment"""

    COL_MSGTYPE = 0
    COL_PCTIME  = 1
    COL_INITTIME  = 2
    COL_FINALTIME = 3
    COL_MSG       = 4
    COL_INFO      = 5

    def __init__(self, session):
        """
        :param session: session reference
        :type session: pycontrolgui.windows.detail.entities.session_window.SessionWindow
        """
        super().__init__(session.name)
        self.session = session

        self._trials = ControlCombo('Trial', changed_event=self.__draw_diagram_evt)
        self._image  = ControlImage('Diagram')

        self.formset = ['_trials', '_image']

        self.__load_trials()
        self.__draw_diagram_evt()

    def __load_trials(self):
        self._trials.clear()

        res = self.session.data.query("TYPE in ['TRIAL']")

        for i, (index, msg) in enumerate( res.iterrows() ):
            
            self._trials.add_item(
                'Trial {0}'.format(i), index
            )
            
   
    def __draw_diagram_evt(self):
        trial_begin = self._trials.value
        trial_end   = None

        if trial_begin:
            trials = self._trials._items
            trials_indexes = sorted(trials.values())

            for trial_index in trials_indexes:
                if trial_index>trial_begin:
                    trial_end = trial_index
                    break

        if trial_end:
            indexes_filter = "index>{0} and index<{1}".format(trial_begin, trial_end)
        else:
            indexes_filter = "index>{0}".format(trial_begin)        

        if trial_end:
            query = "TYPE in ['TRANSITION'] and index>{0} and index<{1}".format(trial_begin, trial_end)
        else:
            query = "TYPE in ['TRANSITION'] and index>{0}".format(trial_begin)

        res = self.session.data.query("TYPE in ['EVENT-SUMMARY']").query(indexes_filter)
        events = []
        for _, msg in res.iterrows():
            events.append( (msg[self.COL_INITTIME], msg[self.COL_INFO]) )
        
        res = self.session.data.query("TYPE in ['TRANSITION']").query(indexes_filter)

        # this time, in graph_type we specify we want a DIrected GRAPH
        graph = pydot.Dot(graph_type='digraph', font='verdana')

        nodes  = {}
        states = res.MSG.unique()
        for state in states:
            node = pydot.Node(state, color='#ff4000')
            graph.add_node(node)
            nodes[state] = node

        prev = None
        curr = None

        curr_event = 0
        for i, (index, msg) in enumerate( res.iterrows() ):
            state = msg[self.COL_MSG]
            begin = msg[self.COL_INITTIME]

            if math.isnan(begin): continue

            curr  = nodes[state]

            for evt_i in range(curr_event, len(events) ):
                evt_end = events[evt_i][0]
                if evt_end<=begin:
                    evt_label  = events[evt_i][1]
                    curr_event = evt_i + 1
                    break

            if prev and curr:
                label = '   {0} ({1})'.format(evt_label, evt_end)
                graph.add_edge( pydot.Edge(prev, curr, label=label, fontsize="10.0", arrowhead='normal') )

            prev = curr

        
        graph.write_png('stm-diagram.png',prog='dot')

        # this is too good to be true!

        self._image.value = 'stm-diagram.png'



    def show(self, detached = False):
        if self.session.is_running and self.session.setup.detached:
            return
        
        # Prevent the call to be recursive because of the mdi_area
        if not detached:
            if hasattr(self, '_show_called'):
                BaseWidget.show(self)
                return
            self._show_called = True
            self.mainwindow.mdi_area += self
            del self._show_called
        else:
            BaseWidget.show(self)
        
        self._stop = False # flag used to close the gui in the middle of a loading
        

    def before_close_event(self):
        self.session.stmdiagram_action.setEnabled(True)






    @property
    def mainwindow(self):
        return self.session.mainwindow

    @property
    def title(self):
        return BaseWidget.title.fget(self)

    @title.setter
    def title(self, value):
        title = 'Stm: {0}'.format(value)
        BaseWidget.title.fset(self, title)
